import inspect
from collections import defaultdict

import pytest
from _pytest.unittest import UnitTestCase

__version__ = '2.0.0'


def pytest_collection_modifyitems(session, config, items):
    errors = defaultdict(list)

    for item in items:
        parent = getattr(item, 'parent', None)
        if (
            parent is None or
            not isinstance(parent, UnitTestCase) or
            parent in errors
        ):
            continue

        klass = parent.cls

        for name in ('setUpClass', 'setUpTestData', 'setUp', 'tearDown', 'tearDownClass'):
            try:
                klass.__dict__[name]
            except KeyError:
                # Not defined in this class (may be in superclass)
                continue

            func = getattr(klass, name)  # Fetch *bound* method

            real_func = get_real_func(func)

            # Unwrap any decorators, we only care about inspecting the innermost
            while hasattr(real_func, '__wrapped__'):
                real_func = get_real_func(real_func.__wrapped__)

            if 'super' not in real_func.__code__.co_names:
                errors[parent].append(name)

    if errors:
        raise pytest.UsageError(*[
            error_msg(p, names) for p, names in errors.items()
        ])


def get_real_func(func):
    """
    Copied from patchy.

    Duplicates some of the logic implicit in inspect.getsource(). Basically
    some function-esque things, such as classmethods, aren't functions but we
    can peel back the layers to the underlying function very easily.
    """
    if inspect.ismethod(func):
        return func.__func__
    else:
        return func


def error_msg(parent, names):
    return '{parent_id} does not call super() in {names}'.format(
        parent_id=parent.nodeid,
        names=', '.join(names),
    )



import re

from setuptools import setup


def get_version(filename):
    with open(filename, 'r') as fp:
        contents = fp.read()
    return re.search(r"__version__ = ['\"]([^'\"]+)['\"]", contents).group(1)


version = get_version('pytest_super_check.py')

with open('README.rst', 'r') as readme_file:
    readme = readme_file.read()

with open('HISTORY.rst', 'r') as history_file:
    history = history_file.read().replace('.. :changelog:', '')


setup(
    name='pytest-super-check',
    version=version,
    description='Pytest plugin to check your TestCase classes call super in setUp, tearDown, etc.',
    long_description=readme + '\n\n' + history,
    author="Adam Johnson",
    author_email='me@adamj.eu',
    url='https://github.com/adamchainz/pytest-super-check',
    py_modules=['pytest_super_check'],
    include_package_data=True,
    install_requires=[
        'pytest',
    ],
    python_requires='>=3.4',
    license="BSD",
    zip_safe=False,
    keywords='pytest, super, unittest, testcase',
    entry_points={
        'pytest11': ['super_check = pytest_super_check'],
    },
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Framework :: Pytest',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ],
)



pytest_plugins = ['pytester']


def test_it_does_not_complain_when_everything_supers_correctly(testdir):
    testdir.makepyfile(
        test_one="""
        from unittest import TestCase


        class MyTests(TestCase):

            @classmethod
            def setUpClass(cls):
                super(MyTests, cls).setUpClass()

            @classmethod
            def setUpTestData(cls):
                super(MyTests, cls).setUpTestData()

            def setUp(self):
                super(MyTests, self).setUp()

            def test_one(self):
                pass

            def tearDown(self):
                super(MyTests, self).tearDown()

            @classmethod
            def tearDownClass(cls):
                super(MyTests, cls).tearDownClass()
        """
    )
    out = testdir.runpytest()
    out.assert_outcomes(passed=1, failed=0)


def test_it_complains_when_a_case_does_not_super_in_setUp(testdir):
    testdir.makepyfile(
        test_one="""
        from unittest import TestCase


        class MyTests(TestCase):

            def setUp(self):
                self.x = 1

            def test_one(self):
                pass
        """
    )
    out = testdir.runpytest()
    assert out.ret > 0
    out.stderr.fnmatch_lines([
        'ERROR: test_one.py::MyTests does not call super() in setUp'
    ])


def test_it_complains_when_a_case_does_not_super_in_tearDown(testdir):
    testdir.makepyfile(
        test_one="""
        from unittest import TestCase


        class MyTests(TestCase):

            def tearDown(self):
                self.x = 1

            def test_one(self):
                pass
        """
    )
    out = testdir.runpytest()
    assert out.ret > 0
    out.stderr.fnmatch_lines([
        'ERROR: test_one.py::MyTests does not call super() in tearDown'
    ])


def test_it_complains_when_a_case_does_not_super_in_setUpClass(testdir):
    testdir.makepyfile(
        test_one="""
        from unittest import TestCase


        class MyTests(TestCase):

            @classmethod
            def setUpClass(cls):
                cls.x = 1

            def test_one(self):
                pass
        """
    )
    out = testdir.runpytest()
    assert out.ret > 0
    out.stderr.fnmatch_lines([
        'ERROR: test_one.py::MyTests does not call super() in setUpClass'
    ])


def test_it_complains_when_a_case_does_not_super_in_setUpTestData(testdir):
    testdir.makepyfile(
        test_one="""
        from unittest import TestCase

        class TestData(object):  # Fake for Django test case
            @classmethod
            def setUpTestData(cls):
                pass


        class MyTests(TestData, TestCase):

            @classmethod
            def setUpTestData(cls):
                cls.x = 1

            def test_one(self):
                pass
        """
    )
    out = testdir.runpytest()
    assert out.ret > 0
    out.stderr.fnmatch_lines([
        'ERROR: test_one.py::MyTests does not call super() in setUpTestData'
    ])


def test_it_complains_when_a_case_does_not_super_in_tearDownClass(testdir):
    testdir.makepyfile(
        test_one="""
        from unittest import TestCase


        class MyTests(TestCase):

            @classmethod
            def tearDownClass(cls):
                cls.x = 1

            def test_one(self):
                pass
        """
    )
    out = testdir.runpytest()
    assert out.ret > 0
    out.stderr.fnmatch_lines([
        'ERROR: test_one.py::MyTests does not call super() in tearDownClass'
    ])


def test_it_complains_when_a_case_does_not_super_in_setUp_and_setUpClass(testdir):
    testdir.makepyfile(
        test_one="""
        from unittest import TestCase


        class MyTests(TestCase):

            @classmethod
            def setUpClass(cls):
                cls.x = 1

            def setUp(self):
                self.y = 1

            def test_one(self):
                pass
        """
    )
    out = testdir.runpytest()
    assert out.ret > 0
    out.stderr.fnmatch_lines([
        'ERROR: test_one.py::MyTests does not call super() in setUpClass, setUp'
    ])


def test_it_does_not_complain_when_a_decorator_is_used_but_super_is_called(testdir):
    testdir.makepyfile(
        test_one="""
        from functools import wraps
        from unittest import TestCase


        def mydecorator(func):
            @wraps(func)
            def wrapper(self):
                return func(self)
            wrapper.__wrapped__ = func  # Python 2.7 compat
            return wrapper


        class MyTests(TestCase):

            @mydecorator
            def setUp(self):
                super(MyTests, self).setUp()

            def test_one(self):
                pass
        """
    )
    out = testdir.runpytest()
    out.assert_outcomes(passed=1, failed=0)


def test_it_complains_when_a_decorator_is_used_and_super_is_not_called(testdir):
    testdir.makepyfile(
        test_one="""
        from functools import wraps
        from unittest import TestCase


        def mydecorator(func):
            @wraps(func)
            def wrapper(self):
                return func(self)
            wrapper.__wrapped__ = func  # Python 2.7 compat
            return wrapper


        class MyTests(TestCase):

            @mydecorator
            def setUp(self):
                self.x = 1

            def test_one(self):
                pass
        """
    )
    out = testdir.runpytest()
    assert out.ret > 0
    out.stderr.fnmatch_lines([
        'ERROR: test_one.py::MyTests does not call super() in setUp'
    ])
